/*
 * Copyright (c) 2012 Christiano F. Haesbaert <haesbaert@openbsd.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <err.h>
#include <errno.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "test.h"

#define RAND(x) arc4random_uniform(x)

void test_sched_get_priority_minmax(void);
void test_sched_rr_get_interval(void);
void test_sched_setgetscheduler(void);
void test_setsched_getparam_minmax(pid_t, int);
void test_sched_setgetparam(void);
void spin(void);

void
test_sched_get_priority_minmax(void)
{
	/* Test 1: min is 0 */
	CHECKe(sched_get_priority_min(SCHED_OTHER));
	CHECKe(sched_get_priority_min(SCHED_RR));
	CHECKe(sched_get_priority_min(SCHED_FIFO));
	/* Test 2: max is 0 for OTHER, 31 for RR and FIFO */
	CHECKe(sched_get_priority_max(SCHED_OTHER));
	CHECKi(sched_get_priority_max(SCHED_RR), 31);
	CHECKi(sched_get_priority_max(SCHED_FIFO), 31);
	/* Test 3: Check for EINVAL */
	CHECKerr(sched_get_priority_min(0xff));
	ASSERT(errno == EINVAL);
	CHECKerr(sched_get_priority_max(0xff));
	ASSERT(errno == EINVAL);
}

void
test_sched_rr_get_interval(void)
{
	struct timespec tp;
	
	/* Test 1: Ourselves. */
	CHECKe(sched_rr_get_interval(0, &tp));
	/* Test 2: Check an existing pid */
	CHECKe(sched_rr_get_interval(getppid(), &tp));
	/* Test 3: Check an unexisting pid, nondeterminisc ! */
	CHECKerr(sched_rr_get_interval((pid_t) -1, &tp));
	ASSERT(errno == ESRCH);
	/* TODO check tp value */
}

void
test_sched_setgetscheduler(void)
{
	struct sched_param sp;
	/* Test 1: Basic getscheduler */
	CHECKi(sched_getscheduler(0), SCHED_OTHER);
	CHECKi(sched_getscheduler(getpid()), SCHED_OTHER);
	CHECKi(sched_getscheduler(getppid()), SCHED_OTHER);
	/* Test 2: Check for ESRCH, nondeterministic */
	CHECKerr(sched_getscheduler((pid_t) -1));
	ASSERT(errno == ESRCH);
	/* Test 3: Set and test */
	bzero(&sp, sizeof(sp));
	sp.sched_priority = sched_get_priority_min(SCHED_RR);
	CHECKe(sched_setscheduler(0, SCHED_RR, &sp));
	spin();
	CHECKi(sched_getscheduler(0), SCHED_RR);
	sp.sched_priority = sched_get_priority_max(SCHED_RR);
	CHECKe(sched_setscheduler(0, SCHED_RR, &sp));
	spin();
	CHECKi(sched_getscheduler(0), SCHED_RR);
	sp.sched_priority = sched_get_priority_max(SCHED_FIFO);
	CHECKe(sched_setscheduler(0, SCHED_FIFO, &sp));
	CHECKi(sched_getscheduler(0), SCHED_FIFO);
	sp.sched_priority = sched_get_priority_max(SCHED_OTHER);
	CHECKe(sched_setscheduler(0, SCHED_OTHER, &sp));
	spin();
	CHECKi(sched_getscheduler(0), SCHED_OTHER);
	/* Test 4: Check min and max */
	test_setsched_getparam_minmax(getpid(), SCHED_FIFO);
	test_setsched_getparam_minmax(getpid(), SCHED_RR);
	test_setsched_getparam_minmax(getpid(), SCHED_OTHER);
	/* Test 5: Change other process sched */
	test_setsched_getparam_minmax(getppid(), SCHED_FIFO);
	test_setsched_getparam_minmax(getppid(), SCHED_RR);
	test_setsched_getparam_minmax(getppid(), SCHED_OTHER);
}

void
test_setsched_getparam_minmax(pid_t pid, int sched)
{
	struct sched_param sp;

	bzero(&sp, sizeof(sp));
	sp.sched_priority = sched_get_priority_min(sched);
	CHECKe(sched_setscheduler(pid, sched, &sp));
	CHECKi(sched_getscheduler(pid), sched);
	CHECKe(sched_getparam(pid, &sp));
	printf("peguei %d\n", sp.sched_priority);
	ASSERT(sp.sched_priority == sched_get_priority_min(sched));

	sp.sched_priority = sched_get_priority_max(sched);
	CHECKe(sched_setscheduler(pid, sched, &sp));
	CHECKi(sched_getscheduler(pid), sched);
	CHECKe(sched_getparam(pid, &sp));
	ASSERT(sp.sched_priority == sched_get_priority_max(sched));
}

void
test_sched_setgetparam(void)
{
	struct sched_param sp;

	/* Test 1: Make sure param is set */
	bzero(&sp, sizeof(sp));
	sp.sched_priority = sched_get_priority_min(SCHED_RR);
	CHECKe(sched_setscheduler(0, SCHED_RR, &sp));
	sp.sched_priority = 5;
	CHECKe(sched_setparam(0, &sp));
	bzero(&sp, sizeof(sp));
	CHECKe(sched_getparam(0, &sp));
	ASSERT(sp.sched_priority == 5);
	/* Test 2: Same thing on FIFO */
	bzero(&sp, sizeof(sp));
	sp.sched_priority = sched_get_priority_min(SCHED_FIFO);
	CHECKe(sched_setscheduler(0, SCHED_FIFO, &sp));
	sp.sched_priority = 5;
	CHECKe(sched_setparam(0, &sp));
	bzero(&sp, sizeof(sp));
	CHECKe(sched_getparam(0, &sp));
	ASSERT(sp.sched_priority == 5);
	/* Test 3: Check for EINVAL */
	bzero(&sp, sizeof(sp));
	sp.sched_priority = sched_get_priority_min(SCHED_OTHER);
	CHECKe(sched_setscheduler(0, SCHED_OTHER, &sp));
	spin();
	sp.sched_priority = sched_get_priority_min(SCHED_OTHER) + 1;
	CHECKerr(sched_setparam(0, &sp));
	/* Test 4: Getparam on SCHED_OTHER must return 0 */
	sp.sched_priority = 5;
	CHECKe(sched_getparam(0, &sp));
	ASSERT(sp.sched_priority == 0);
}

/*
 * A naive attempt of providing some entropy
 */
void
spin(void)
{
	u_int32_t x;

	printf("sleeping...\n");
	sleep(RAND(3));

	printf("spinning...\n");
	for (x = RAND(5) * 100000000; x; x--)
		;
	if (RAND(2)) {
		printf("yielding...\n");
		sched_yield();
	}
}

int
main(void)
{

	test_sched_get_priority_minmax();
	test_sched_rr_get_interval();
	test_sched_setgetscheduler();
	test_sched_setgetparam();

	SUCCEED;
}
